"""
biodome
=======

Controlled environments.

   Copyright 2018 Caleb Hattingh

   Licensed under the Apache License, Version 2.0 (the "License");
   you may not use this file except in compliance with the License.
   You may obtain a copy of the License at

       http://www.apache.org/licenses/LICENSE-2.0

   Unless required by applicable law or agreed to in writing, software
   distributed under the License is distributed on an "AS IS" BASIS,
   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   See the License for the specific language governing permissions and
   limitations under the License.

"""
import ast
import contextlib
import errno
import functools
import logging
import os
import typing

try:
    # Python 3
    from collections import UserDict  # pragma: no cover
except ImportError:  # pragma: no cover
    # Python 2
    from UserDict import IterableUserDict as UserDict  # pragma: no cover

if typing.TYPE_CHECKING:  # pragma: no cover
    from typing import Callable, Any


__version__ = "2021.10.1"
logger = logging.getLogger(__name__)


def biodome(name, default=None, cast=None):
    # type: (str, Any, Callable) -> Any
    raw_value = os.environ.get(name)
    if raw_value is None:
        return default

    if default is None and cast is None:
        # No type information. Just pass through.
        return raw_value

    raw_value = raw_value.strip()

    # Use the same type as default as the cast
    type_ = cast or type(default)

    if bool in (cast, type_):
        return raw_value.lower() in (
            "1",
            "y",
            "yes",
            "on",
            "active",
            "activated",
            "enabled",
            "true",
            "t",
            "ok",
            "yeah",
        )

    try:
        if type_ in (dict, list, set, tuple):
            raw_value = ast.literal_eval(raw_value)
            return (type(raw_value) == type_ and raw_value) or default
        return type_(raw_value)
    except:
        logger.error(
            'Env var %s: cast "%s" to type %s failed. The default will be' "used.",
            name,
            raw_value,
            str(type_),
        )
        return default


class _Environ(UserDict):
    def __init__(self, *args, **kwargs):
        self.data = os.environ

    def get(self, key, default=None, cast=None):
        # type: (str, Any, Callable) -> Callable
        return biodome(key, default, cast)

    def get_callable(self, key, default=None, cast=None):
        # type: (str, Any, Callable) -> Callable[[None], None]
        return functools.partial(self.get, key, default=default, cast=cast)

    def __setitem__(self, key, value):
        os.environ[key] = str(value)


environ = _Environ()


@contextlib.contextmanager
def env_change(name, value):
    """Context manager to temporarily change the value of an env var."""
    # TODO: move this upstream to the biodome package
    if name in environ:
        old = environ[name]

        def reset():
            environ[name] = old

    else:

        def reset():
            del environ[name]

    try:
        environ[name] = value
        yield
    finally:
        reset()


def load_env_file(path, raises=False):
    # type: (str, bool) -> None
    """
    Load an file which specifies the values of environment variables. An
    example:

        # This sets the log level for all the loggers in the program
        LOGGER_LEVEL=info

        # Hourly backups are stored at this path and named with a timestamp.
        BACKUP_PATH=/data/backups/

        # The number of times to retry outgoing HTTP requests if the status
        # code is > 500
        RETRY_TIME=5


    The name of the environment variable must be on the left and the value
    on the right. Each variable must be on its own line. Lines starting with
    a # are considered comments and are ignored.

    :raises - If true, this method with raise if there is no file at the
        specified path. If false, the method will return having done nothing.
    """
    try:
        with open(path) as f:
            for line in f:
                line = line.strip()
                if not line or line[0] == "#":
                    continue
                name, _, value = line.partition("=")
                name = name.strip()
                value = value.strip()
                environ[name] = value
    except IOError as e:
        # Python 3 raises a FileNotFound and python 2 an IOError. So we can
        # check the error number to see if it was a missing file.
        if e.errno != errno.ENOENT or raises:
            raise



import os
import sys

from copy import deepcopy
from uuid import uuid4

import pytest
import biodome


PY2 = sys.version_info < (3, 0)


def test_missing():
    assert biodome.biodome('BLAH') is None


def test_missing_default():
    assert biodome.biodome('BLAH', None) is None


@pytest.mark.parametrize('name,default,setting,result', [
    ('X', 2, '123', 123),
    ('X', '2', '123', '123'),
    ('X', True, '123', False),
    ('X', True, '', False),
    ('X', True, '0', False),
    ('X', True, 't', True),
    ('X', True, '1', True),
    ('X', True, 'y', True),
    ('X', True, 'yes', True),
    ('X', True, 'YES', True),
    ('X', True, 'Yes', True),
    ('X', True, 'Y', True),
    ('X', True, 'True', True),
    ('X', True, 'true', True),
    ('X', True, 'TRUE', True),
    ('X', True, 'on', True),
    ('X', True, 'On', True),
    ('X', True, 'ON', True),
    ('X', True, 'active', True),
    ('X', True, 'ACTIVE', True),
    ('X', True, 'Active', True),
    ('X', False, 'y', True),
    ('X', False, 'yes', True),
    ('X', False, 'YES', True),
    ('X', False, 'Yes', True),
    ('X', False, 'Y', True),
    ('X', False, 'True', True),
    ('X', False, 'true', True),
    ('X', False, 'TRUE', True),
    ('X', False, 'on', True),
    ('X', False, 'On', True),
    ('X', False, 'ON', True),
    ('X', False, 'active', True),
    ('X', False, 'ACTIVE', True),
    ('X', False, 'Active', True),

    ('X', False, 'yas', False),
    ('X', False, 'blah', False),
    ('X', False, '123', False),
    ('X', False, '____', False),
    ('X', False, 'heyy', False),

    ('X', 123, 'heyy', 123),
    ('X', -123, '-123', -123),
    ('X', 0, '0', 0),

    ('X', 0.1, '0', 0),
    ('X', 0.1, '1.2', 1.2),
    ('X', -0.1, '-1.2', -1.2),
    ('X', 1, '1.0', 1),

    ('X', 'heya', 'blah', 'blah'),

    ('X', {}, '{"a": 1}', dict(a=1)),
    ('X', {}, '{"a": "blah", "b": "bleh"}', dict(a='blah', b='bleh')),

    ('X', [], '[1, 2, 3]', [1, 2, 3]),
    ('X', [], '[]', []),
    ('X', [], '[', []),
    ('X', [], '[blah]', []),
    ('X', [], '["blah"]', ['blah']),

    ('X', (1, 2), 'blah', (1, 2)),
    ('X', (), 'blah', ()),
    ('X', (), '(1, 2)', (1, 2)),
    ('X', (), '(1,2)', (1, 2)),
    ('X', (), '(   1   ,   2   )', (1, 2)),
    ('X', (), '(1)', ()),
    ('X', (), '(1,)', (1,)),

    ('X', '0', '1.053', '1.053'),
])
def test_param_set(name, default, setting, result):
    os.environ[name] = str(setting)
    assert biodome.biodome(name, default) == result


@pytest.mark.skipif(PY2, reason="requires Python 3+")
@pytest.mark.parametrize('name,default,setting,result', [
    ('X', {1, 2, 3}, '{1, 2}', {1, 2}),
    ('X', {1, 2, 3}, '{"1":2}', {1, 2, 3}),
])
def test_param(name, default, setting, result):
    """ast.literal_eval is only supported in Python 3"""
    os.environ[name] = str(setting)
    assert biodome.biodome(name, default) == result


@pytest.mark.parametrize('name,cast,setting,result', [
    ('X', bool, 'blah', False),
    ('X', bool, '1', True),
    ('X', bool, '0', False),
    ('X', bool, 'True', True),
    ('X', bool, 'False', False),
    ('X', bool, 'true', True),
    ('X', bool, 'false', False),

    ('X', float, 'blah', None),
    ('X', float, '1', 1.0),
    ('X', float, '1', 1),
    ('X', float, '1e-17', 1e-17),

    ('X', int, 'blah', None),
    ('X', int, '1', 1),
    ('X', int, '1.0', None),

    ('X', str, '1.0', '1.0'),
    ('X', str, '1', '1'),

    ('X', list, '1', None),
    ('X', list, '[1]', [1]),

    ('X', dict, '{"a": 123}', dict(a=123)),
])
def test_cast(name, cast, setting, result):
    os.environ[name] = str(setting)
    assert biodome.biodome(name, cast=cast) == result


@pytest.mark.skipif(PY2, reason="requires Python 3+")
@pytest.mark.parametrize('name,cast,setting,result', [
    ('X', set, '{2}', {2}),
])
def test_cast_set(name, cast, setting, result):
    os.environ[name] = str(setting)
    assert biodome.biodome(name, cast=cast) == result


@pytest.mark.parametrize('name,default,setting,result', [
    ('X', [], 'list(1,2,3)', [1, 2, 3]),
    ('X', set(), 'set([2])', {2}),
    ('X', {}, 'dict(a=1)', {"a": 1}),
])
def test_noeval(name, default, setting, result):
    """Verify that no code evaluation occurs"""
    os.environ[name] = str(setting)
    assert biodome.biodome(name, default=default) != result
    assert biodome.biodome(name, default=default) == default


def test_environ_set():
    assert 'blah' not in os.environ
    biodome.environ['blah'] = '123'
    assert 'blah' in os.environ
    assert os.environ.get('blah') == '123'
    del biodome.environ['blah']
    assert 'blah' not in os.environ


def test_environ_int():
    assert not os.environ.get('blah')
    biodome.environ['blah'] = 123

    # Both ways work
    assert os.environ['blah'] == '123'
    assert biodome.environ['blah'] == '123'

    # NOTE: we get an int out because the default is type int
    assert biodome.environ.get('blah', default=0) == 123

    del biodome.environ['blah']
    assert 'blah' not in os.environ


def test_environ_types():
    assert not os.environ.get('blah')
    biodome.environ['blah'] = dict(a=[1, 2, 3])

    rep = "{'a': [1, 2, 3]}"

    # Both ways work - old and new
    assert os.environ['blah'] == rep
    assert biodome.environ['blah'] == rep

    # NOTE: we get an int out because the default is type int
    assert biodome.environ.get('blah', default={}) == dict(a=[1, 2, 3])

    del biodome.environ['blah']
    assert 'blah' not in os.environ


def test_callable():
    assert not os.environ.get('MY_SETTING')
    biodome.environ['MY_SETTING'] = dict(a=[1, 2, 3])
    MY_SETTING = biodome.environ.get_callable('MY_SETTING', {})
    assert MY_SETTING() == dict(a=[1, 2, 3])


def test_callable_int():
    os.environ['MY_SETTING2'] = '5'
    MY_SETTING2 = biodome.environ.get('MY_SETTING2')
    assert MY_SETTING2 == '5'


def test_env_changer_new():
    assert 'BLAH' not in biodome.environ

    with biodome.env_change('BLAH', 123):
        assert biodome.environ.get('BLAH', 0) == 123

    assert 'BLAH' not in biodome.environ


def test_env_changer_existing():
    biodome.environ['BLAH'] = 456

    with biodome.env_change('BLAH', 123):
        assert biodome.environ.get('BLAH', 0) == 123

    assert biodome.environ.get('BLAH', 0) == 456
    del biodome.environ['BLAH']


def test_loading_file(tmpdir):
    p = tmpdir.join(str(uuid4()) + '.env')
    p.write_text(u'# This is a comment\nX_SET=123', 'utf8')
    expected = deepcopy(biodome.environ.data)
    expected['X_SET'] = '123'

    biodome.load_env_file(str(p))
    actual = biodome.environ.data
    assert actual == expected
    assert biodome.environ.get('X_SET', 1) == 123


def test_loading_missing_file(tmpdir):
    p = str(tmpdir) + str(uuid4()) + '.env'
    before = deepcopy(biodome.environ.data)
    biodome.load_env_file(str(p))
    assert biodome.environ.data == before


def test_loading_missing_file_raises(tmpdir):
    p = str(tmpdir) + str(uuid4()) + '.env'
    before = deepcopy(biodome.environ.data)
    with pytest.raises(IOError):
        biodome.load_env_file(str(p), raises=True)
    assert biodome.environ.data == before
